@@ -4,11 +4,13 @@
 # NOTE: Argument 'filter' conflicts with built-in 'filter' function
 _filter = filter
 
-from _libldap import _LDAPError, _LDAPObject
+from _libldap import _LDAPError, _LDAPObject, _LDAPObjectControl
+from .constants import LDAP_CONTROL_PAGEDRESULTS
 from collections import OrderedDict as _OrderedDict
 
 __all__ = (
     'LDAP',
+    'LDAPControl',
     'LDAPError',
 )
 
@@ -22,9 +24,17 @@ class LDAPError(Exception):
     def __init__(self, message, return_code, *args, **kwargs):
         self.message = message
         self.return_code = return_code
+        self.args = args
+        for key, value in kwargs.items():
+            setattr(self, key, value)
+
+    def __repr__(self):
+        additional_info = ' %s' % (getattr(self, 'ppolicy_msg', ''),)
+        return 'LDAPError(%s (%d)%s)' % (self.message, self.return_code, additional_info)
 
     def __str__(self):
-        return '%s (%d)' % (self.message, self.return_code)
+        additional_info = ' %s' % (getattr(self, 'ppolicy_msg', ''),)
+        return '%s (%d)%s' % (self.message, self.return_code, additional_info)
 
 
 class _OrderedEntry(_OrderedDict):
@@ -57,12 +67,14 @@ class LDAP(_LDAPObject):
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
 
-    def bind(self, who, password, async=False):
+    def bind(self, who, password, controls=None, async=False):
         """
         Parameters
         ----------
         who : str
         password : str
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
         async : bool
             If True, return result immediately
             (the default is False, which means operation will
@@ -79,11 +91,14 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            msgid = super().bind(who, password)
+            if controls is not None:
+                msgid = super().bind(who, password, controls)
+            else:
+                msgid = super().bind(who, password)
             if async:
                 # Not set bind_user
                 return msgid
-            result = super().result(msgid)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
@@ -114,6 +129,7 @@ class LDAP(_LDAPObject):
                attrsonly=False,
                timeout=0,
                sizelimit=0,
+               controls=None,
                ordered_attributes=False,
                async=False):
         """
@@ -135,6 +151,8 @@ class LDAP(_LDAPObject):
             (the default is 0, which implies unlimited)
         sizelimit : int, optional
             (the default is 0, which implies unlimited)
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
         ordered_attributes : bool, optional
             If ordered_attributes is True, the order of the attributes in entry
             are remembered (the default is False).
@@ -153,16 +171,44 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            msgid = super().search(base, scope, filter, attributes,
-                                   int(attrsonly), timeout, sizelimit)
+            if controls is not None:
+                msgid = super().search(base, scope, filter, attributes,
+                                       int(attrsonly), timeout, sizelimit, controls)
+            else:
+                msgid = super().search(base, scope, filter, attributes,
+                                       int(attrsonly), timeout, sizelimit)
             if async:
                 return msgid
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
-        return self.search_result(msgid, timeout=timeout,
+        return self.search_result(msgid, timeout=timeout, controls=controls,
                                   ordered_attributes=ordered_attributes)
 
-    def add(self, dn, attributes, async=False):
+    def paged_search(self,
+               base,
+               scope=0x0000,
+               filter='(objectClass=*)',
+               attributes=None,
+               attrsonly=False,
+               timeout=0,
+               sizelimit=0,
+               pagesize=100,
+               ordered_attributes=False):
+        _pagesize = ('%d' % (pagesize,)).encode('utf-8')
+        controls = _LDAPObjectControl()
+        controls.add_control(LDAP_CONTROL_PAGEDRESULTS, _pagesize, False)
+        initial = True
+        while initial or controls.get_pr_cookie() is not None:
+            initial = False
+            try:
+                msgid = super().search(base, scope, filter, attributes,
+                                       int(attrsonly), timeout, sizelimit, controls)
+            except _LDAPError as e:
+                raise LDAPError(str(e), LDAP_ERROR) from None
+            yield from self.search_result(msgid, timeout=timeout, controls=controls,
+                                          ordered_attributes=ordered_attributes)
+
+    def add(self, dn, attributes, controls=None, async=False):
         """
         Parameters
         ----------
@@ -171,6 +217,8 @@ class LDAP(_LDAPObject):
             List of tuple. tuple has two items:
                 attr   - Attribute name
                 values - List of value
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
         async : bool
             If True, return result immediately
             (the default is False, which means operation will
@@ -187,16 +235,19 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            msgid = super().add(dn, attributes)
+            if controls is not None:
+                msgid = super().add(dn, attributes, controls)
+            else:
+                msgid = super().add(dn, attributes)
             if async:
                 return msgid
-            result = super().result(msgid)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
             raise LDAPError(**result)
 
-    def modify(self, dn, changes, async=False):
+    def modify(self, dn, changes, controls=None, async=False):
         """
         Parameters
         ----------
@@ -206,6 +257,8 @@ class LDAP(_LDAPObject):
                 attr   - Attribute name
                 values - List of value
                 mod_op - Modify operation (e.g.: LDAP_MOD_REPLACE)
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
         async : bool
             If True, return result immediately
             (the default is False, which means operation will
@@ -222,20 +275,25 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            msgid = super().modify(dn, changes)
+            if controls is not None:
+                msgid = super().modify(dn, changes, controls)
+            else:
+                msgid = super().modify(dn, changes)
             if async:
                 return msgid
-            result = super().result(msgid)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
             raise LDAPError(**result)
 
-    def delete(self, dn, async=False):
+    def delete(self, dn, controls=None, async=False):
         """
         Parameters
         ----------
         dn : str
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
         async : bool
             If True, return result immediately
             (the default is False, which means operation will
@@ -252,16 +310,19 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            msgid = super().delete(dn)
+            if controls is not None:
+                msgid = super().delete(dn, controls)
+            else:
+                msgid = super().delete(dn)
             if async:
                 return msgid
-            result = super().result(msgid)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
             raise LDAPError(**result)
 
-    def rename(self, dn, newrdn, newparent, deleteoldrdn=True, async=False):
+    def rename(self, dn, newrdn, newparent, deleteoldrdn=True, controls=None, async=False):
         """
         Parameters
         ----------
@@ -270,6 +331,8 @@ class LDAP(_LDAPObject):
         newparent : str
         deleteoldrdn : bool
             (the default is True, which means oldrdn is deleted after renamed)
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
         async : bool
             If True, return result immediately
             (the default is False, which means operation will
@@ -286,22 +349,27 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            msgid = super().rename(dn, newrdn, newparent, int(deleteoldrdn))
+            if controls is not None:
+                msgid = super().rename(dn, newrdn, newparent, int(deleteoldrdn), controls)
+            else:
+                msgid = super().rename(dn, newrdn, newparent, int(deleteoldrdn))
             if async:
                 return msgid
-            result = super().result(msgid)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
             raise LDAPError(**result)
 
-    def compare(self, dn, attribute, value):
+    def compare(self, dn, attribute, value, controls=None):
         """
         Parameters
         ----------
         dn : str
         attribute : str
         value : str
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
 
         Returns
         -------
@@ -316,8 +384,11 @@ class LDAP(_LDAPObject):
         This method operates synchronously.
         """
         try:
-            msgid = super().compare(dn, attribute, value)
-            result = super().result(msgid)
+            if controls is not None:
+                msgid = super().compare(dn, attribute, value, controls)
+            else:
+                msgid = super().compare(dn, attribute, value)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] == LDAP_COMPARE_TRUE:
@@ -327,8 +398,13 @@ class LDAP(_LDAPObject):
         else:
             raise LDAPError(**result)
 
-    def whoami(self):
+    def whoami(self, controls=None):
         """
+        Parameters
+        ----------
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
+
         Returns
         -------
         str
@@ -343,8 +419,11 @@ class LDAP(_LDAPObject):
         This method operates synchronously.
         """
         try:
-            msgid = super().whoami()
-            result = super().result(msgid)
+            if controls is not None:
+                msgid = super().whoami(controls)
+            else:
+                msgid = super().whoami()
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
@@ -354,13 +433,15 @@ class LDAP(_LDAPObject):
         else:
             return 'anonymous'
 
-    def passwd(self, user, oldpw=None, newpw=None):
+    def passwd(self, user, oldpw=None, newpw=None, controls=None):
         """
         Parameters
         ----------
         user : str
         oldpw : str, optional
         newpw : str, optional
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
 
         Returns
         -------
@@ -376,8 +457,11 @@ class LDAP(_LDAPObject):
         This method operates synchronously.
         """
         try:
-            msgid = super().passwd(user, oldpw, newpw)
-            result = super().result(msgid)
+            if controls is not None:
+                msgid = super().passwd(user, oldpw, newpw, controls)
+            else:
+                msgid = super().passwd(user, oldpw, newpw)
+            result = self.result(msgid, controls=controls)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
         if result['return_code'] != LDAP_SUCCESS:
@@ -388,8 +472,13 @@ class LDAP(_LDAPObject):
         else:
             return newpw
 
-    def start_tls(self):
+    def start_tls(self, controls=None):
         """
+        Parameters
+        ----------
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
+
         Returns
         -------
         None
@@ -400,7 +489,10 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            super().start_tls()
+            if controls is not None:
+                super().start_tls(controls)
+            else:
+                super().start_tls()
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
 
@@ -480,11 +572,13 @@ class LDAP(_LDAPObject):
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
 
-    def abandon(self, msgid):
+    def abandon(self, msgid, controls=None):
         """
         Parameters
         ----------
         msgid : int
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
 
         Returns
         -------
@@ -496,15 +590,20 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            return super().abandon(msgid)
+            if controls is not None:
+                return super().abandon(msgid, controls)
+            else:
+                return super().abandon(msgid)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
 
-    def cancel(self, msgid):
+    def cancel(self, msgid, controls=None):
         """
         Parameters
         ----------
         msgid : int
+        controls : LDAPControl, optional
+            (the default is None, which implies no controls are set)
 
         Returns
         -------
@@ -516,11 +615,14 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            return super().cancel(msgid)
+            if controls is not None:
+                return super().cancel(msgid, controls)
+            else:
+                return super().cancel(msgid)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
 
-    def result(self, msgid, all=True, timeout=3):
+    def result(self, msgid, all=True, timeout=3, controls=None):
         """
         Parameters
         ----------
@@ -539,7 +641,10 @@ class LDAP(_LDAPObject):
         LDAPError
         """
         try:
-            return super().result(msgid, int(all), timeout)
+            if controls is not None:
+                return super().result(msgid, int(all), timeout, controls)
+            else:
+                return super().result(msgid, int(all), timeout)
         except _LDAPError as e:
             raise LDAPError(str(e), LDAP_ERROR) from None
 
@@ -568,13 +673,19 @@ class LDAP(_LDAPObject):
             ordered_attributes = kwargs.pop('ordered_attributes')
         else:
             ordered_attributes = False
+        results = self.result(*args, **kwargs)
+        if results:
+            if results[-1]['return_code'] != LDAP_SUCCESS:
+                raise LDAPError(**results[-1])
         if ordered_attributes:
             return [_OrderedEntry([(key, entry[key])
                                    for key in entry['__order__']])
-                    for entry in self.result(*args, **kwargs)
-                    if '__order__' in entry]
+                    for entry in results if '__order__' in entry]
         else:
             return [dict([(key, value) for key, value in entry.items()
                           if key != '__order__'])
-                    for entry in self.result(*args, **kwargs)
-                    if '__order__' in entry]
+                    for entry in results if '__order__' in entry]
+
+
+class LDAPControl(_LDAPObjectControl):
+    pass
