@@ -15,7 +15,10 @@ LDAPObject_rename(LDAPObject *self, PyObject *args)
 	const char *newrdn;
 	const char *newparent;
 	int deleteoldrdn;
+	PyObject *controls = NULL;
+	LDAPObjectControl *ldapoc = NULL;
 	LDAPControl **sctrls = NULL;
+	LDAPControl **cctrls = NULL;
 	int rc;
 	int msgid;
 
@@ -24,11 +27,18 @@ LDAPObject_rename(LDAPObject *self, PyObject *args)
 		return NULL;
 	}
 
-	if (!PyArg_ParseTuple(args, "sssi", &dn, &newrdn, &newparent, &deleteoldrdn))
+	if (!PyArg_ParseTuple(args, "sssi|O!", &dn, &newrdn, &newparent, &deleteoldrdn,
+				&LDAPObjectControlType, &controls))
 		return NULL;
 
+	if (controls) {
+		ldapoc = (LDAPObjectControl *)controls;
+		sctrls = ldapoc->sctrls;
+		cctrls = ldapoc->cctrls;
+	}
+
 	LDAP_BEGIN_ALLOW_THREADS
-	rc = ldap_rename(self->ldap, dn, newrdn, newparent, deleteoldrdn, sctrls, NULL, &msgid);
+	rc = ldap_rename(self->ldap, dn, newrdn, newparent, deleteoldrdn, sctrls, cctrls, &msgid);
 	LDAP_END_ALLOW_THREADS
 	if (rc != LDAP_SUCCESS) {
 		PyErr_SetString(LDAPError, ldap_err2string(rc));
